#include "stdafx.h"
#include "LeoHelpers_String.h"
#include "LeoHelpers_Scope.h"
#include "LeoHelpers_Opus.h"
#include "LeoHelpers_Misc.h"
#include "LeoHelpers_Debug.h"
#include "LeoHelpers_Window.h"
#include "LeoHelpers_Layout.h"

void LeoHelpers::CenterWindow(HWND hWndToMove, HWND hWndRelative)
{
	bool bResult = true;

	RECT rectSnap;

	RECT rectRelative;

	if (NULL != hWndRelative)
	{
		::GetWindowRect(hWndRelative, &rectRelative);

		if (!LeoHelpers::GetAreasFromWindow(NULL, &rectSnap, hWndRelative, MONITOR_DEFAULTTONEAREST))
		{
			bResult = false;
		}
	}
	else
	{
		// Get work area under the mouse pointer.

		POINT pointCursor;
		MONITORINFO monitorInfo;
		ZeroMemory(&monitorInfo, sizeof(monitorInfo));
		monitorInfo.cbSize = sizeof(monitorInfo);
		HMONITOR hMonitor;

		if (GetCursorPos(&pointCursor)
		&&	(NULL != (hMonitor = ::MonitorFromPoint(pointCursor, MONITOR_DEFAULTTOPRIMARY)))
		&&	::GetMonitorInfo(hMonitor, &monitorInfo))
		{
			rectRelative = monitorInfo.rcWork;
			CopyRect(&rectSnap, &rectRelative);
		}
		else if (::GetWindowRect(::GetDesktopWindow(), &rectRelative))
		{
			CopyRect(&rectSnap, &rectRelative);
		}
		else
		{
			bResult = false;
		}
	}

	RECT rectToMove;

	if (bResult && ::GetWindowRect(hWndToMove, &rectToMove))
	{
		int relativeWidth  = (rectRelative.right  - rectRelative.left);
		int relativeHeight = (rectRelative.bottom - rectRelative.top);
		int toMoveWidth    = (rectToMove.right    - rectToMove.left);
		int toMoveHeight   = (rectToMove.bottom   - rectToMove.top);

		int x = rectRelative.left + ( (relativeWidth  - toMoveWidth ) / 2 );
		int y = rectRelative.top  + ( (relativeHeight - toMoveHeight) / 2 );

		if (x < rectSnap.left)
		{
			x = rectSnap.left;
		}

		if ((x + toMoveWidth) > rectSnap.right)
		{
			x = rectSnap.right - toMoveWidth;
		}

		if (y < rectSnap.top)
		{
			y = rectSnap.top;
		}

		if ((y + toMoveHeight) > rectSnap.bottom)
		{
			y = rectSnap.bottom - toMoveHeight;
		}

		::SetWindowPos(hWndToMove, NULL, x, y, 0, 0, SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE);
	}
}

bool LeoHelpers::FitToWorkspace(HWND hWnd)
{
	// If the window isn't already on a monitor then we'll use the primary display;
	// however, code which relies on that is probably buggy and calling us before the
	// window has been positions, so assert in those cases.
	assert(NULL != ::MonitorFromWindow(hWnd, MONITOR_DEFAULTTONULL));

	RECT rcWindow = {0};
	RECT rcWork   = {0};
	MINMAXINFO mmi = {0};

	::SendMessage(hWnd, WM_GETMINMAXINFO, 0, reinterpret_cast< LPARAM >( &mmi ));

	if (!::GetWindowRect(hWnd, &rcWindow)
	||	!LeoHelpers::GetAreasFromWindow(NULL, &rcWork, hWnd, MONITOR_DEFAULTTOPRIMARY)
	||	mmi.ptMinTrackSize.x == 0
	||	mmi.ptMinTrackSize.y == 0)
	{
		return false; // did not resize
	}

	LONG lWorkWidth    = rcWork.right    - rcWork.left;
	LONG lWorkHeight   = rcWork.bottom   - rcWork.top;
	LONG lWindowWidth  = rcWindow.right  - rcWindow.left;
	LONG lWindowHeight = rcWindow.bottom - rcWindow.top;

	if (lWindowWidth  <= lWorkWidth
	&&	lWindowHeight <= lWorkHeight)
	{
		return false; // did not resize
	}

	bool bChanged = false;

	if (lWindowWidth > lWorkWidth
	&&	lWindowWidth > mmi.ptMinTrackSize.x)
	{
		lWindowWidth = max(lWorkWidth, mmi.ptMinTrackSize.x);
		rcWindow.right = rcWindow.left + lWindowWidth;
		bChanged = true;
	}

	if (lWindowHeight > lWorkHeight
	&&	lWindowHeight > mmi.ptMinTrackSize.y)
	{
		lWindowHeight = max(lWorkHeight, mmi.ptMinTrackSize.y);
		rcWindow.bottom = rcWindow.top + lWindowHeight;
		bChanged = true;
	}

	if (rcWindow.right > rcWork.right)
	{
		rcWindow.right = rcWork.right;
		rcWindow.left  = rcWindow.right - lWindowWidth;
		bChanged = true;
	}

	if (rcWindow.left < rcWork.left)
	{
		rcWindow.left  = rcWork.left;
		rcWindow.right = rcWindow.left + lWindowWidth;
		bChanged = true;
	}

	if (rcWindow.bottom > rcWork.bottom)
	{
		rcWindow.bottom = rcWork.bottom;
		rcWindow.top    = rcWindow.bottom - lWindowHeight;
		bChanged = true;
	}

	if (rcWindow.top < rcWork.top)
	{
		rcWindow.top    = rcWork.top;
		rcWindow.bottom = rcWindow.top + lWindowHeight;
		bChanged = true;
	}

	if (bChanged)
	{
		LeoHelpers::MoveWindowRect(hWnd, &rcWindow, TRUE);
	}

	return bChanged;
}

bool LeoHelpers::GetAreasFromWindow(RECT *pRectFull, RECT *pRectWork, HWND hWnd, DWORD dwFlags)
{
	if (NULL != pRectFull) { ZeroMemory(pRectFull, sizeof(RECT)); }
	if (NULL != pRectWork) { ZeroMemory(pRectWork, sizeof(RECT)); }

	bool bResult = false;

	bool bGetDefault = true;

	HMONITOR hMonitor = ::MonitorFromWindow(hWnd, dwFlags);

	if ((NULL == hMonitor) && (MONITOR_DEFAULTTONULL == dwFlags))
	{
		bGetDefault = false;
	}
	else if (NULL != hMonitor)
	{
		MONITORINFO mi;
		ZeroMemory(&mi, sizeof(mi));
		mi.cbSize = sizeof(mi);

		if (::GetMonitorInfo(hMonitor, &mi)
		&&	(NULL == pRectFull || CopyRect(pRectFull, &mi.rcMonitor))
		&&	(NULL == pRectWork || CopyRect(pRectWork, &mi.rcWork   )))
		{
			bResult = true;
		}
	}

	if (bGetDefault && (!bResult))
	{
		bResult = true;

		if (bResult && NULL != pRectFull)
		{
			pRectFull->left   = 0;
			pRectFull->top    = 0;
			pRectFull->right  = ::GetSystemMetrics(SM_CXFULLSCREEN);
			pRectFull->bottom = ::GetSystemMetrics(SM_CYFULLSCREEN);

			if (0 == pRectFull->right || 0 == pRectFull->bottom)
			{
				bResult = false;
			}
		}

		if (bResult && NULL != pRectWork)
		{
			if (!::SystemParametersInfo(SPI_GETWORKAREA, 0, pRectWork, 0))
			{
				bResult = false;
			}
		}
	}

	return(bResult);
}

/*
// DeleteObject the result
HFONT LeoHelpers::GetMessageFont()
{
	HFONT hFont = NULL;

	NONCLIENTMETRICS ncm = {0};
	ncm.cbSize = sizeof(NONCLIENTMETRICS);

	if (::SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(NONCLIENTMETRICS), &ncm, 0))
	{
		//ncm.lfMessageFont.lfHeight *= 2;
		//ncm.lfMessageFont.lfWidth *= 2;

		hFont = ::CreateFontIndirect(&ncm.lfMessageFont);
	}

	return hFont;
}
*/

LeoHelpers::DialogLayoutHelper::DialogLayoutHelper(HWND hwndDlg)
: m_hwndDlg(hwndDlg)
, m_lBaseX(0)
, m_lBaseY(0)
{
	/*
	// Work out the dialog units from the font the dialog is actually using.
	// See: http://support.microsoft.com/kb/145994
	// See also: http://blogs.msdn.com/michkap/archive/2007/08/16/4421328.aspx (looks like this is sensible even for non-Latin fonts)

	hFont = NULL;

	if (hFont == NULL)
	{
		// If no font is passed in then we might as well fall back on the default ::MapDialogRect.
		// There's no point calling WM_GETFONT on the dialog window as dialog windows seem to completely
		// ignore WM_SETFONT and just return the dialog template font, which is what ::MapDialogRect uses.
		// However, we'll assert in debug builds as we should avoid triggering this fallback.
		//assert(hFont != NULL);
		//hFont = reinterpret_cast< HFONT >(::SendMessage(hwndDlg, WM_GETFONT, 0, 0));
	}

	if (hFont != NULL)
	{
		HDC hdc = ::GetDC(hwndDlg);

		if (hdc != NULL)
		{
			HGDIOBJ hFontOld = ::SelectObject(hdc, hFont);

			TEXTMETRIC tm = {0};
			SIZE size = {0};

			if (::GetTextMetrics(hdc, &tm)
			&&	::GetTextExtentPoint32(hdc, L"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", 52, &size))
			{
				m_lBaseX = (size.cx / 26 + 1) / 2;
				m_lBaseY = tm.tmHeight;
			}

			::SelectObject(hdc, hFontOld);

			::ReleaseDC(hwndDlg, hdc);
		}
	}
	*/

	m_rect1.left   = 4; // Button X and related-controls X
	m_rect1.right  = 9; // Checkbox box width.
	m_rect1.top    = 4; // Button Y and related-controls Y
	m_rect1.bottom = 7; // Margin (X and Y)
	this->MapDialogRectProperly(&m_rect1);

	m_rect2.left   = 2; // Checkbox box-to-text gap width.
	m_rect2.right  = 0;
	m_rect2.top    = 3; // Label Assoc Gap Y
	m_rect2.bottom = 8; // Checkbox box height.
	this->MapDialogRectProperly(&m_rect2);

	m_rect3.left   = 50; // Standard button width.
	m_rect3.right  = 0;
	m_rect3.top    = 11; // First control in group-box.
	m_rect3.bottom = 14; // Edit control height. (Single line edit control.) Also button height.
	this->MapDialogRectProperly(&m_rect3);

	m_rect4.left   = 0;
	m_rect4.right  = 0;
	m_rect4.top    = 9;  // Progress control height.
	m_rect4.bottom = 0;
	this->MapDialogRectProperly(&m_rect4);
}

/*
// Font stuff has been ditched in favour of re-writing the dialog resource to use the font we want
// and being able to use the ::MapDialogRect without having to calculate things ourselves.
BOOL LeoHelpers::DialogLayoutHelper::MapDialogRectProperly(RECT *pRect) const
{
	if (m_lBaseX > 0 && m_lBaseY > 0)
	{
		pRect->left   = (pRect->left   * m_lBaseX) / 4;
		pRect->right  = (pRect->right  * m_lBaseX) / 4;
		pRect->top    = (pRect->top    * m_lBaseY) / 8;
		pRect->bottom = (pRect->bottom * m_lBaseY) / 8;

		return TRUE;
	}

	// If we failed to calculate the base units in the constructor, fall back on the Win32 API.
	return ::MapDialogRect(m_hwndDlg, pRect);
}
*/

bool LeoHelpers::DialogLayoutHelper::MeasureTextSingleLine(HWND hWndControl, const wchar_t *szText, RECT *prc, bool bNoPrefix)
{
	bool bResult = false;

	::SetRectEmpty(prc);

	HDC hDC = ::GetDC(hWndControl);

	if (hDC != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(hWndControl, WM_GETFONT, 0, 0));
		HGDIOBJ hOldFont = NULL;

		if (hFont != NULL)
		{
			hOldFont = ::SelectObject(hDC, hFont);
		}

		UINT uFormat = DT_CALCRECT | DT_NOCLIP | DT_SINGLELINE;

		if (bNoPrefix)
		{
			uFormat |= DT_NOPREFIX;
		}

		if (hFont == NULL)
		{
			uFormat |= DT_INTERNAL; // Use system font.
		}

		if (0 != ::DrawText(hDC, szText, static_cast< int >( wcslen(szText) ), prc, uFormat))
		{
			bResult = true;
		}

		if (hFont != NULL)
		{
			::SelectObject(hDC, hOldFont);
		}

		::ReleaseDC(hWndControl, hDC);
	}

	return bResult;
}

LONG LeoHelpers::DialogLayoutHelper::GetComboBoxItemWidth(HWND hWndCombo)
{
	COMBOBOXINFO comboInfo = {0};
	comboInfo.cbSize = sizeof(comboInfo);

	if (::GetComboBoxInfo(hWndCombo, &comboInfo))
	{
		LONG lWidth = comboInfo.rcItem.right - comboInfo.rcItem.left;
		if (lWidth >= 0)
		{
			return lWidth;
		}
	}

	return 0;
}

// Calling this (or otherwise moving/sizing the combo) will zero the drop-down list's height on XP.
// See EnsureComboBoxListHeight, below, for more info and how to fix it.
bool LeoHelpers::DialogLayoutHelper::EnsureComboBoxWidth(HWND hWndCombo, LONG lMaxLabelWidth, bool bResizeWindow, RECT *prcComboOut) const
{
	// Add on a fudge-factor. Using the real widths seems to result in slightly clipped text (though maybe just the anti-aliasing).
	// It looks nicer to have a bit of a gap between the text and the drop-down anyway.
	// (This fudge-factor is needed for the GetComboBoxItemWidth comparison; not sure if also needed for the drop-down box
	// comparison. Either way, it should be added to the max width and not subtracted from the combo width to avoid going negative/zero.)
	lMaxLabelWidth += GetRelatedControlGapXPixels();

	RECT rcCombo = {0};
	if (::GetWindowRect(hWndCombo, &rcCombo))
	{
		// Get minimum of space for label in the control itself and in the drop-down list.

		LONG lControlLabelWidth = GetComboBoxItemWidth(hWndCombo);
		LONG lDropDownLabelWidth = LeoHelpers::WidthRect(rcCombo) - (GetSystemMetrics(SM_CXVSCROLL) + GetSystemMetrics(SM_CXEDGE) + GetSystemMetrics(SM_CXFIXEDFRAME) * 2);

		LONG lComboCurrent = lControlLabelWidth;
		if (lComboCurrent > lDropDownLabelWidth && lDropDownLabelWidth > 0)
		{
			lComboCurrent = lDropDownLabelWidth;
		}

		if (lComboCurrent > 0)
		{
			if (lMaxLabelWidth <= lComboCurrent)
			{
				if (prcComboOut != NULL) { *prcComboOut = rcCombo; }
				return true; // Already wide enough.
			}

			rcCombo.right += (lMaxLabelWidth - lComboCurrent);
			
			if (!bResizeWindow || LeoHelpers::MoveWindowRect(hWndCombo, &rcCombo, TRUE))
			{
				if (prcComboOut != NULL) { *prcComboOut = rcCombo; }
				return true;
			}
		}
	}

	if (prcComboOut != NULL) { ::SetRectEmpty(prcComboOut); }
	return false;
}

// MUST be called on Windows XP, else the list will have zero height. (Vista and above are more intelligent.)
// MUST be called *AFTER* any other moving or sizing of the combo, else the stupid thing gets undone on XP.
// (If you get the combo control's rect it'll tell you the rect of the actual control. If you set the rect
// then the height defines that of the drop-down list (NOT the control!). So just getting the rect, adjusting
// the position and/or width (not even touching the height!) and then setting the rect will make the drop-down
// zero height on XP. Sigh of bloody sighs.)
bool LeoHelpers::DialogLayoutHelper::EnsureComboBoxListHeight(HWND hWndCombo, int iMinLineCountOrZero) const
{
	if (iMinLineCountOrZero < 1)
	{
		LRESULT lrCount = ::SendMessage(hWndCombo, CB_GETCOUNT, 0, 0);

		leoAssert(lrCount >= 0);
		if (lrCount >= 0)
		{
			iMinLineCountOrZero = static_cast< int >( lrCount );
		}
	}

	if (iMinLineCountOrZero < 1)
	{
		iMinLineCountOrZero = 5; // Some nominal size so the drop-down isn't invisible.
	}

	if (!::SendMessage(hWndCombo, CB_SETMINVISIBLE, iMinLineCountOrZero, 0))
	{
		leoAssert(false);
	//	return false; -- Might as well continue even if that failed for some reason.
	}

	return true;
}

bool LeoHelpers::DialogLayoutHelper::ResizeStatic(HWND hwndStatic, RECT *pInOutRect, bool bResize, bool bRedraw, bool bSingleLine) const
{
	bool bResult = false;

	RECT r = {0};

	std::wstring strWindowText;

	if (LeoHelpers::GetWindowTextString(hwndStatic, &strWindowText, false))
	{
		HDC hDC = ::GetDC(hwndStatic);

		if (hDC != NULL)
		{
			HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(hwndStatic, WM_GETFONT, 0, 0));
			HGDIOBJ hOldFont = NULL;

			if (hFont != NULL)
			{
				hOldFont = ::SelectObject(hDC, hFont);
			}

			UINT uFormat = DT_CALCRECT | DT_NOCLIP | DT_EXPANDTABS; // I think DT_EXPANDTABS is always in place with statics. Not sure though.

			if (bSingleLine)
			{
				uFormat |= DT_SINGLELINE;
			}
			else if (pInOutRect != NULL)
			{
				uFormat |= DT_WORDBREAK;
				r = *pInOutRect;
			}

			if (SS_NOPREFIX == (::GetWindowLongPtr(hwndStatic, GWL_STYLE) & SS_NOPREFIX))
			{
				uFormat |= DT_NOPREFIX;
			}

			if (SS_ENDELLIPSIS == (::GetWindowLongPtr(hwndStatic, GWL_STYLE) & SS_ENDELLIPSIS))
			{
				uFormat |= DT_END_ELLIPSIS;
			}

			if (SS_WORDELLIPSIS == (::GetWindowLongPtr(hwndStatic, GWL_STYLE) & SS_WORDELLIPSIS))
			{
				uFormat |= DT_WORD_ELLIPSIS;
			}

			if (SS_PATHELLIPSIS == (::GetWindowLongPtr(hwndStatic, GWL_STYLE) & SS_PATHELLIPSIS))
			{
				// DrawText with DT_PATH_ELLIPSIS has bugs which mean, for some strings, it draws something wider than the
				// specified rect, resulting in either clipping or (if DT_NOCLIP given) painting outside the rect.
				//
				// I'm not sure what that means for this ResizeStatic code, though. Does DrawText still calculate the
				// correct required rect, or does it get it wrong the same as it mis-calculates while painting/truncating?
				//
				// If a fix is needed, this is how we fixed actually drawing (not calc-rec'ing) text with DT_PATH_ELLIPSIS in Opus:
				//
				//int FixedDrawText(HDC hDC,LPWSTR pszText,int iLen,RECT *prc,DWORD dwFlags)
				//{
				//	const size_t PCP_SLOP = 5; // Extra chars just in case PathCompactPath docs are wrong and it may add to an already-MAX_PATH-length path.
				//	wchar_t tchBuf[MAX_PATH+1+PCP_SLOP];
				//	if ((dwFlags&DT_PATH_ELLIPSIS) && iLen!=0)
				//	{
				//		int iSize;
				//		if (iLen==-1)
				//		{
				//			iSize=lstrlen(pszText)+1; // We know there's room for a null.
				//			StringCchCopy(tchBuf,_countof(tchBuf)-PCP_SLOP,pszText);
				//		}
				//		else
				//		{
				//			iSize=iLen; // We don't know if there's room for a null.
				//			StringCchCopyN(tchBuf,_countof(tchBuf)-PCP_SLOP,pszText,iLen);
				//		}
				//		PathCompactPath(hDC,tchBuf,prc->right-prc->left);
				//		if (dwFlags&DT_MODIFYSTRING)
				//		{
				//			// DT_MODIFYSTRING is documented as requiring 4 extra chars' space in the buffer.
				//			// Assumption: PathCompactPath has the same requirement.
				//			// If iLen isn't -1 and the modified string uses all 4 extra chars (maybe the elipsis
				//			// is longer in other languages?) then it seems there's a choice between three options:
				//			// 1) Assume there's room for a 5th char (the null), which is very bad if there isn't.
				//			// 2) Write the four chars and don't null-terminate the string, which is very bad if the caller expects a null.
				//			// 3) Write three chars and then a null, truncating the string.
				//			// I've chosen #3 as it's unlikely to happen and it's better to truncate a string than crash!
				//			StringCchCopy(pszText,iSize+4,tchBuf);
				//		}
				//		else
				//		{
				//			pszText=tchBuf;
				//		}
				//		iLen=-1;
				//		dwFlags&=~(DT_MODIFYSTRING|DT_PATH_ELLIPSIS);
				//	}
				//	return DrawText(hDC,pszText,iLen,prc,dwFlags);
				//}
				//
				uFormat |= DT_PATH_ELLIPSIS;
			}

			if (hFont == NULL)
			{
				uFormat |= DT_INTERNAL; // Use system font.
			}

			if (0 != ::DrawText(hDC, strWindowText.c_str(), static_cast< int >(strWindowText.length()), &r, uFormat))
			{
				bResult = true;
			}

			if (hFont != NULL)
			{
				::SelectObject(hDC, hOldFont);
			}

			::ReleaseDC(hwndStatic, hDC);
		}
	}

	if (bResult)
	{
		if (bResize)
		{
			::SetWindowPos(hwndStatic, NULL, 0, 0, r.right - r.left, r.bottom - r.top,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER | (bRedraw ? 0 : SWP_NOREDRAW));
		}

		if (pInOutRect != NULL)
		{
			*pInOutRect = r;
		}
	}

	return bResult;
}

bool LeoHelpers::DialogLayoutHelper::ResizeProgress(HWND hwndProgress, RECT *pOutRect, bool bResize, bool bRedraw) const
{
	bool bResult = false;

	RECT r = {0};
	r.bottom = m_rect4.top;

	bResult = true; // That's all we do right now.

	if (bResult)
	{
		if (bResize)
		{
			::SetWindowPos(hwndProgress, NULL, 0, 0, r.right - r.left, r.bottom - r.top,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER | (bRedraw ? 0 : SWP_NOREDRAW));
		}

		if (pOutRect != NULL)
		{
			*pOutRect = r;
		}
	}

	return bResult;
}

bool LeoHelpers::DialogLayoutHelper::ResizeCheckbox(HWND hwndCheckbox, RECT *pOutRect, bool bResize, bool bRedraw) const
{
	bool bResult = false;

	RECT r = {0};

	std::wstring strWindowText;

	if (LeoHelpers::GetWindowTextString(hwndCheckbox, &strWindowText, false))
	{
		HDC hDC = ::GetDC(hwndCheckbox);

		if (hDC != NULL)
		{
			HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(hwndCheckbox, WM_GETFONT, 0, 0));
			HGDIOBJ hOldFont = NULL;

			if (hFont != NULL)
			{
				hOldFont = ::SelectObject(hDC, hFont);
			}

			UINT uFormat = DT_CALCRECT | DT_NOCLIP | DT_SINGLELINE;

		//	if (SS_NOPREFIX == (::GetWindowLongPtr(hwndStatic, GWL_STYLE) & SS_NOPREFIX))
		//	{
		//		uFormat |= DT_NOPREFIX;
		//	}

			if (hFont == NULL)
			{
				uFormat |= DT_INTERNAL; // Use system font.
			}

			if (0 != ::DrawText(hDC, strWindowText.c_str(), static_cast< int >(strWindowText.length()), &r, uFormat))
			{
				r.right += (m_rect1.right + m_rect2.left);

				if (r.bottom < m_rect2.bottom)
				{
					r.bottom = m_rect2.bottom;
				}

				bResult = true;
			}

			if (hFont != NULL)
			{
				::SelectObject(hDC, hOldFont);
			}

			::ReleaseDC(hwndCheckbox, hDC);
		}
	}

	if (bResult)
	{
		if (bResize)
		{
			::SetWindowPos(hwndCheckbox, NULL, 0, 0, r.right - r.left, r.bottom - r.top,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER | (bRedraw ? 0 : SWP_NOREDRAW));
		}

		if (pOutRect != NULL)
		{
			*pOutRect = r;
		}
	}

	return bResult;
}

bool LeoHelpers::DialogLayoutHelper::ResizeEdit(HWND hwndEdit, RECT *pOutRect, bool bResize, bool bRedraw) const
{
	bool bResult = false;

	RECT r = {0};

	r.bottom = m_rect3.bottom;
	r.right  = m_rect3.bottom; // Arbitrary width.

	bResult = true;

	if (bResult)
	{
		if (bResize)
		{
			::SetWindowPos(hwndEdit, NULL, 0, 0, r.right - r.left, r.bottom - r.top,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER | (bRedraw ? 0 : SWP_NOREDRAW));
		}

		if (pOutRect != NULL)
		{
			*pOutRect = r;
		}
	}

	return bResult;
}

bool LeoHelpers::DialogLayoutHelper::ResizeButton(HWND hwndButton, RECT *pOutRect, bool bResize, bool bRedraw, bool bIgnoreMinimumWidth) const
{
	// This is also called on ComboBox controls.
	// (Does not resize ComboBoxes for their content, however. Use MeasureTextSingleLine and EnsureComboBoxWidth *after* ResizeButton.)

	bool bResult = false;

	RECT r = {0};

	std::wstring strWindowText;

	if (LeoHelpers::GetWindowTextString(hwndButton, &strWindowText, false))
	{
		HDC hDC = ::GetDC(hwndButton);

		if (hDC != NULL)
		{
			HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(hwndButton, WM_GETFONT, 0, 0));
			HGDIOBJ hOldFont = NULL;

			if (hFont != NULL)
			{
				hOldFont = ::SelectObject(hDC, hFont);
			}

			UINT uFormat = DT_CALCRECT | DT_NOCLIP | DT_SINGLELINE;

		//	if (SS_NOPREFIX == (::GetWindowLongPtr(hwndStatic, GWL_STYLE) & SS_NOPREFIX))
		//	{
		//		uFormat |= DT_NOPREFIX;
		//	}

			if (hFont == NULL)
			{
				uFormat |= DT_INTERNAL; // Use system font.
			}

			if (0 != ::DrawText(hDC, strWindowText.c_str(), static_cast< int >(strWindowText.length()), &r, uFormat))
			{
				if (bIgnoreMinimumWidth) // Intended for "..." browse buttons.
				{
					r.right += (m_rect1.left * 3); // Use 1.5 * button-gap as the margin. Note that the UX Guide doesn't specify an official margin for button text.
				}
				else
				{
					r.right += (m_rect1.left * 4); // Use 2 * button-gap as the margin. Note that the UX Guide doesn't specify an official margin for button text.

					if ((r.right - r.left) < m_rect3.left)
					{
						r.right = r.left + m_rect3.left; // Minimum standard button width.
					}
				}

				if (r.bottom < m_rect3.bottom)
				{
					r.bottom = m_rect3.bottom; // Minimum standard button height.
				}

				bResult = true;
			}

			if (hFont != NULL)
			{
				::SelectObject(hDC, hOldFont);
			}

			::ReleaseDC(hwndButton, hDC);
		}
	}

	if (bResult)
	{
		if (bResize)
		{
			::SetWindowPos(hwndButton, NULL, 0, 0, r.right - r.left, r.bottom - r.top,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER | (bRedraw ? 0 : SWP_NOREDRAW));
		}

		if (pOutRect != NULL)
		{
			*pOutRect = r;
		}
	}

	return bResult;

}

void LeoHelpers::DialogLayoutHelper::SetGroupBoxText(HWND hwndGroupBox, const wchar_t *szText, bool bNoPrefix) const
{
	bool bResult = false;

	RECT r = {0};

	wchar_t *szTextCopy = NULL;

	if (szText != NULL && szText[0] != L'\0')
	{
		size_t slen = 0;

		for (const wchar_t *szIn = szText; *szIn != L'\0'; ++szIn)
		{
			++slen;
			if (bNoPrefix && *szIn == L'&')
			{
				++slen;
			}
		}

		szTextCopy = new(std::nothrow) wchar_t[slen + 1 + 4]; // + 1 for null, + 4 for DT_MODIFYSTRING

		if (szTextCopy != NULL)
		{
			wchar_t *szOut = szTextCopy;
			for (const wchar_t *szIn = szText; *szIn != L'\0'; ++szIn)
			{
				*szOut++ = *szIn;
				if (bNoPrefix && *szIn == L'&')
				{
					*szOut++ = L'&';
				}
			}
			*szOut = L'\0';
			szOut = NULL;

			RECT r;

			if (::GetClientRect(hwndGroupBox, &r))
			{
				r.right -= 2 * GetMarginPixels();

				HDC hDC = ::GetDC(hwndGroupBox);

				if (hDC != NULL)
				{
					HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(hwndGroupBox, WM_GETFONT, 0, 0));
					HGDIOBJ hOldFont = NULL;

					if (hFont != NULL)
					{
						hOldFont = ::SelectObject(hDC, hFont);
					}

					UINT uFormat = DT_CALCRECT | DT_NOCLIP | DT_SINGLELINE | DT_MODIFYSTRING | DT_WORD_ELLIPSIS;

					if (hFont == NULL)
					{
						uFormat |= DT_INTERNAL; // Use system font.
					}

					if (0 == ::DrawText(hDC, szTextCopy, static_cast< int >(slen), &r, uFormat))
					{
						delete[] szTextCopy;
						szTextCopy = NULL;
					}

					if (hFont != NULL)
					{
						::SelectObject(hDC, hOldFont);
					}

					::ReleaseDC(hwndGroupBox, hDC);
				}
			}
		}
	}

	if (szTextCopy != NULL)
	{
		::SetWindowText(hwndGroupBox, szTextCopy);
		delete[] szTextCopy;
	}
	else
	{
		::SetWindowText(hwndGroupBox, szText);
	}
}

bool LeoHelpers::WindowMaker::CreateCheck(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, const wchar_t *szLabel, bool bResize, bool bRadio, bool bTriState, bool bAuto, bool bTabStop) const
{
	DWORD dwStyle = WS_CHILD | WS_VISIBLE;

	if (bTabStop)                dwStyle |= WS_TABSTOP;

	if (bRadio && bAuto)         dwStyle |= BS_AUTORADIOBUTTON;
	else if (bRadio)             dwStyle |= BS_RADIOBUTTON;
	else if (bTriState && bAuto) dwStyle |= BS_AUTO3STATE;
	else if (bTriState)          dwStyle |= BS_3STATE;
	else if (bAuto)              dwStyle |= BS_AUTOCHECKBOX;
	else                         dwStyle |= BS_CHECKBOX;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, WC_BUTTON, szLabel, dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}

		m_pLayout->ResizeCheckbox(hWndResult, pOutRect, bResize, false);
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateEdit(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, bool bResize, LONG lWidth, bool bAutoHScroll, bool bRight, bool bNumber, bool bReadOnly, bool bPassword) const
{
	DWORD dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP;
	if (bAutoHScroll) dwStyle |= ES_AUTOHSCROLL;
	if (bRight)       dwStyle |= ES_RIGHT;
	if (bNumber)      dwStyle |= ES_NUMBER;
	if (bReadOnly)    dwStyle |= ES_READONLY;
	if (bPassword)    dwStyle |= ES_PASSWORD;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY | WS_EX_CLIENTEDGE, WC_EDIT, L"", dwStyle,
		0, 0, lWidth, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}

		m_pLayout->ResizeEdit(hWndResult, pOutRect, bResize, false);
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateSpin(HWND *pOutHwnd, ULONG uiCtrlId, bool bSetBuddyInt, bool bRight, bool bArrowKeys, bool bNoThousands) const
{
	DWORD dwStyle = WS_CHILD | WS_VISIBLE;
	if (bSetBuddyInt) dwStyle |= UDS_SETBUDDYINT;
	if (bRight)       dwStyle |= UDS_ALIGNRIGHT;
	if (bArrowKeys)   dwStyle |= UDS_ARROWKEYS;
	if (bNoThousands) dwStyle |= UDS_NOTHOUSANDS;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, UPDOWN_CLASS, L"", dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateGroup(HWND *pOutHwnd, ULONG uiCtrlId, const wchar_t *szLabel) const
{
	const DWORD dwStyle = WS_CLIPSIBLINGS | WS_CHILD | WS_VISIBLE | BS_GROUPBOX;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, WC_BUTTON, szLabel, dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateButton(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, const wchar_t *szLabel, bool bResize, bool bDefPushButton) const
{
	DWORD dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_TEXT;
	if (bDefPushButton) dwStyle |= BS_DEFPUSHBUTTON;
	else                dwStyle |= BS_PUSHBUTTON;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, WC_BUTTON, szLabel, dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}

		m_pLayout->ResizeButton(hWndResult, pOutRect, bResize, false, false);
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateListBox(HWND *pOutHwnd, ULONG uiCtrlId, bool bSort, bool bExtendedSel) const
{
	DWORD dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | WS_VSCROLL | LBS_NOINTEGRALHEIGHT | LBS_NOTIFY;
	if (bSort)        dwStyle |= LBS_SORT;
	if (bExtendedSel) dwStyle |= LBS_EXTENDEDSEL;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY | WS_EX_CLIENTEDGE, WC_LISTBOX, L"", dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateListViewReport(HWND *pOutHwnd, ULONG uiCtrlId,
	bool bSortAscending, bool bSortDescending,
	bool bNoColumnHeader, bool bNoSortHeader,
	bool bShowSelAlways, bool bSingleSel, bool bFullRowSelect,
	bool bEditLabels, bool bCheckboxes, bool bShareImageLists) const
{
	DWORD dwStyle = WS_CLIPSIBLINGS | WS_CHILD | WS_VISIBLE | WS_TABSTOP | LVS_REPORT;
	if (bSortAscending)       dwStyle |= LVS_SORTASCENDING;
	else if (bSortDescending) dwStyle |= LVS_SORTDESCENDING;
	if (bNoColumnHeader)      dwStyle |= LVS_NOCOLUMNHEADER;
	else if (bNoSortHeader)   dwStyle |= LVS_NOSORTHEADER;
	if (bShowSelAlways)       dwStyle |= LVS_SHOWSELALWAYS;
	if (bSingleSel)           dwStyle |= LVS_SINGLESEL;
	if (bEditLabels)          dwStyle |= LVS_EDITLABELS;
	if (bShareImageLists)     dwStyle |= LVS_SHAREIMAGELISTS;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY | WS_EX_CLIENTEDGE, WC_LISTVIEW, L"", dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}

		DWORD dwListViewStyle = 0;
		
		if (bFullRowSelect) dwListViewStyle |= LVS_EX_FULLROWSELECT;
		if (bCheckboxes)    dwListViewStyle |= LVS_EX_CHECKBOXES;

		if (LeoHelpers::IsWindowsXPOrAbove())
		{
			dwListViewStyle |= LVS_EX_DOUBLEBUFFER;
		}

		ListView_SetExtendedListViewStyleEx(hWndResult, dwListViewStyle, dwListViewStyle);
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateLabel(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, const wchar_t *szLabel, bool bResize, bool bSingleLine, bool bVCenter) const
{
	DWORD dwStyle = WS_CHILD | WS_VISIBLE;
	if (bVCenter) dwStyle |= SS_CENTERIMAGE;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, WC_STATIC, szLabel, dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}

		m_pLayout->ResizeStatic(hWndResult, pOutRect, bResize, false, bSingleLine);
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateProgress(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, bool bResize, bool bMarquee, bool bSmooth, bool bSmoothReverse) const
{
#ifndef PBS_MARQUEE
const DWORD PBS_MARQUEE = 0x08;
#endif

#ifndef PBS_SMOOTHREVERSE
const DWORD PBS_SMOOTHREVERSE = 0x10;
#endif


	DWORD dwStyle = WS_CHILD | WS_VISIBLE;
	if (bMarquee)       dwStyle |= PBS_MARQUEE;
	if (bSmooth)        dwStyle |= PBS_SMOOTH;
	if (bSmoothReverse) dwStyle |= PBS_SMOOTHREVERSE;
//	if (bVertical)      dwStyle |= PBS_VERTICAL;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, PROGRESS_CLASS, L"", dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		m_pLayout->ResizeProgress(hWndResult, pOutRect, bResize, false);
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateBitmap(HWND *pOutHwnd, ULONG uiCtrlId, bool bNotify) const
{
	DWORD dwStyle = WS_CHILD | WS_VISIBLE | SS_BITMAP | SS_CENTERIMAGE | SS_REALSIZEIMAGE;
	if (bNotify) dwStyle |= SS_NOTIFY;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, WC_STATIC, L"", dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateCombo(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, bool bResize, bool bDropDownList, bool bDropDown) const
{
	DWORD dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | WS_VSCROLL;
	if (bDropDownList)  dwStyle |= CBS_DROPDOWNLIST;
	else if (bDropDown) dwStyle |= CBS_DROPDOWN;
	else                dwStyle |= CBS_SIMPLE;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, WC_COMBOBOX, L"", dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}

		m_pLayout->ResizeButton(hWndResult, pOutRect, bResize, false, false);
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateTab(HWND *pOutHwnd, ULONG uiCtrlId) const
{
	DWORD dwStyle = TCS_FOCUSONBUTTONDOWN | WS_CLIPCHILDREN | WS_CHILD | WS_VISIBLE;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY | WS_EX_CONTROLPARENT, WC_TABCONTROL, L"", dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (hWndResult != NULL)
	{
		HFONT hFont = reinterpret_cast< HFONT >(::SendMessage(m_hWnd, WM_GETFONT, 0, 0));

		if (hFont != NULL)
		{
			::SendMessage(hWndResult, WM_SETFONT, reinterpret_cast< WPARAM >(hFont), FALSE);
		}
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateResizeGrip(HWND *pOutHwnd, ULONG uiCtrlId) const
{
	DWORD dwStyle = WS_CLIPSIBLINGS | WS_CHILD | WS_VISIBLE | SBS_SIZEGRIP | SBS_SIZEBOXBOTTOMRIGHTALIGN;

	HWND hWndResult = ::CreateWindowEx(WS_EX_NOPARENTNOTIFY, WC_SCROLLBAR, L"", dwStyle,
		0, 0, 0, 0, m_hWnd, (HMENU)UlongToPtr(uiCtrlId), m_hModule, NULL);

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

bool LeoHelpers::WindowMaker::CreateTabChildDlg(HWND *pOutHwnd, ULONG uiDlgRes, HWND hWndTab, WPARAM &tabNum, const wchar_t *szLabel, DLGPROC dlgProc, LPARAM lTabParam) const
{
	HWND hWndResult = ::CreateDialogParam(m_hModule, MAKEINTRESOURCE(uiDlgRes), hWndTab, dlgProc, lTabParam);

	if (hWndResult != NULL)
	{
		TCITEM tabItem = {0};
		tabItem.mask = TCIF_PARAM|TCIF_TEXT;
		tabItem.pszText = const_cast< wchar_t * >(szLabel);
		tabItem.lParam = reinterpret_cast< LPARAM >(hWndResult);

		::SendMessage(hWndTab, TCM_INSERTITEM, tabNum++, reinterpret_cast< LPARAM >(&tabItem));
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

#ifdef DOPUS_PLUGIN_HELPER

bool LeoHelpers::OpusWindowMaker::CreateTabChildOpusDlg(HWND *pOutHwnd, ULONG uiDlgRes, HWND hWndTab, WPARAM &tabNum, const wchar_t *szLabel, DLGPROC dlgProc, LPARAM lTabParam, DWORD dwFlags) const
{
	HWND hWndResult = m_pStringLoader->GetOpusPluginHelperUtil()->CreateLangDlgEx(m_hModule, MAKEINTRESOURCE(uiDlgRes), hWndTab, dlgProc, lTabParam, dwFlags);

	if (hWndResult != NULL)
	{
		TCITEM tabItem = {0};
		tabItem.mask = TCIF_PARAM|TCIF_TEXT;
		tabItem.pszText = const_cast< wchar_t * >(szLabel);
		tabItem.lParam = reinterpret_cast< LPARAM >(hWndResult);

		::SendMessage(hWndTab, TCM_INSERTITEM, tabNum++, reinterpret_cast< LPARAM >(&tabItem));
	}

	if (pOutHwnd != NULL)
	{
		*pOutHwnd = hWndResult;
	}

	return (hWndResult != NULL);
}

#endif
