#include "stdafx.h"
#include "LeoHelpers_String.h"
#include "LeoHelpers_Opus.h"

// Opus SDK headers should be included in stdafx.h so their definitions are picked up by both this and LeoHelpers_Opus.h.

#ifndef DOPUS_PLUGIN_LEO_NO_COPYVERSION

#include "LeoHelpers_Version.h"

#ifdef VIEWERPLUGININFO

bool LeoHelpers::CopyVersionResourceToViewerPluginInfo(HMODULE hDllModule, WORD iconId, UINT uiOpusMsgIdName, UINT uiOpusMsgIdDesc, LPVIEWERPLUGININFOW lpVPInfo)
{
	bool bResult = false;

	wchar_t szModName[_MAX_PATH + 1];
	DWORD dwVerSize;
	DWORD dwUnused;

	if (NULL != hDllModule
	&&	GetModuleFileName(hDllModule, szModName, _MAX_PATH)
	&&	(0 != (dwVerSize = GetFileVersionInfoSize(szModName, &dwUnused))))
	{
		BYTE *pVerData = new BYTE[dwVerSize];

		if (GetFileVersionInfo(szModName, NULL, dwVerSize, pVerData))
		{
			LPVOID pBuffer;
			UINT bufLen;

			// Find out language of version resource
			if (VerQueryValue(pVerData, L"\\VarFileInfo\\Translation", &pBuffer, &bufLen)
			&&	4 == bufLen)
			{
				BYTE *pLang = reinterpret_cast<BYTE*>(pBuffer);

				if (LeoHelpers::CopyVersionNumber(&lpVPInfo->dwVersionHigh,  &lpVPInfo->dwVersionLow,      VNT_FILEVERSION,           pVerData)
				&&	LeoHelpers::CopyVersionString(lpVPInfo->lpszCopyright,    lpVPInfo->cchCopyrightMax,   L"LegalCopyright",  pLang, pVerData)
				&&	LeoHelpers::CopyVersionString(lpVPInfo->lpszName,         lpVPInfo->cchNameMax,        L"InternalName",    pLang, pVerData)
				&&	LeoHelpers::CopyVersionString(lpVPInfo->lpszDescription,  lpVPInfo->cchDescriptionMax, L"FileDescription", pLang, pVerData)
				&&	LeoHelpers::CopyVersionString(lpVPInfo->lpszURL,          lpVPInfo->cchURLMax,         L"CompanyName",     pLang, pVerData))
				{
					// If we can, use a translated version of the plugin name.
					// This may fail when the GetString support function is not available, in which case we stick with the English strings.

					DOpusPluginHelperUtil *pHelper = NULL;

					if (uiOpusMsgIdName != 0 && lpVPInfo->lpszName != NULL && lpVPInfo->cchNameMax != 0)
					{
						if (!pHelper) { pHelper = new DOpusPluginHelperUtil; }

						pHelper->GetString(uiOpusMsgIdName, &lpVPInfo->lpszName, lpVPInfo->cchNameMax);
					}

					if (uiOpusMsgIdDesc != 0 && lpVPInfo->lpszDescription != NULL && lpVPInfo->cchDescriptionMax != 0)
					{
						if (!pHelper) { pHelper = new DOpusPluginHelperUtil; }

						pHelper->GetString(uiOpusMsgIdDesc, &lpVPInfo->lpszDescription, lpVPInfo->cchDescriptionMax);
					}

#ifndef UNICODE
					LeoHelpers::StringAppend(lpVPInfo->lpszDescription, " (ANSI)", lpVPInfo->cchDescriptionMax);
#endif
					// If anything fails after the next line then the icon must be destroyed.
					if (0 != iconId && lpVPInfo->cbSize >= VIEWERPLUGININFOW_V4_SIZE)
					{
						lpVPInfo->hIconSmall = reinterpret_cast<HICON>(::LoadImage(hDllModule, MAKEINTRESOURCE(iconId), IMAGE_ICON, 0, 0, LR_DEFAULTCOLOR));
					}

					delete pHelper;

					bResult = true;
				}
			}
		}

		delete [] pVerData;
	}

	return bResult;
}

#endif

#ifdef VFSPLUGININFO

bool LeoHelpers::CopyVersionResourceToVfsPluginInfo(HMODULE hDllModule, WORD iconId, UINT uiOpusMsgIdName, UINT uiOpusMsgIdDesc, LPVFSPLUGININFO lpVPInfo)
{
	bool bResult = false;

	wchar_t szModName[_MAX_PATH + 1];
	DWORD dwVerSize;
	DWORD dwUnused;

	if (NULL != hDllModule
	&&	GetModuleFileName(hDllModule, szModName, _MAX_PATH)
	&&	(0 != (dwVerSize = GetFileVersionInfoSize(szModName, &dwUnused))))
	{
		BYTE *pVerData = new BYTE[dwVerSize];

		if (GetFileVersionInfo(szModName, NULL, dwVerSize, pVerData))
		{
			LPVOID pBuffer;
			UINT bufLen;

			// Find out language of version resource
			if (VerQueryValue(pVerData, L"\\VarFileInfo\\Translation", &pBuffer, &bufLen)
			&&	4 == bufLen)
			{
				BYTE *pLang = reinterpret_cast<BYTE*>(pBuffer);

				if (LeoHelpers::CopyVersionNumber(&lpVPInfo->dwVersionHigh,  &lpVPInfo->dwVersionLow,      VNT_FILEVERSION,           pVerData)
				&&	LeoHelpers::CopyVersionString(lpVPInfo->lpszCopyright,    lpVPInfo->cchCopyrightMax,   L"LegalCopyright",  pLang, pVerData)
				&&	LeoHelpers::CopyVersionString(lpVPInfo->lpszName,         lpVPInfo->cchNameMax,        L"InternalName",    pLang, pVerData)
				&&	LeoHelpers::CopyVersionString(lpVPInfo->lpszDescription,  lpVPInfo->cchDescriptionMax, L"FileDescription", pLang, pVerData)
				&&	LeoHelpers::CopyVersionString(lpVPInfo->lpszURL,          lpVPInfo->cchURLMax,         L"CompanyName",     pLang, pVerData))
				{
					// If we can, use a translated version of the plugin name.
					// This may fail when the GetString support function is not available, in which case we stick with the English strings.

					DOpusPluginHelperUtil *pHelper = NULL;

					if (uiOpusMsgIdName != 0 && lpVPInfo->lpszName != NULL && lpVPInfo->cchNameMax != 0)
					{
						if (!pHelper) { pHelper = new DOpusPluginHelperUtil; }

						pHelper->GetString(uiOpusMsgIdName, &lpVPInfo->lpszName, lpVPInfo->cchNameMax);
					}

					if (uiOpusMsgIdDesc != 0 && lpVPInfo->lpszDescription != NULL && lpVPInfo->cchDescriptionMax != 0)
					{
						if (!pHelper) { pHelper = new DOpusPluginHelperUtil; }

						pHelper->GetString(uiOpusMsgIdDesc, &lpVPInfo->lpszDescription, lpVPInfo->cchDescriptionMax);
					}

#ifndef UNICODE
					LeoHelpers::StringAppend(lpVPInfo->lpszDescription, " (ANSI)", lpVPInfo->cchDescriptionMax);
#endif
					// If anything fails after the next line then the icon must be destroyed.
					if (0 != iconId && lpVPInfo->cbSize >= VFSPLUGININFO_V2_SIZE)
					{
						lpVPInfo->hIconSmall = reinterpret_cast<HICON>(::LoadImage(hDllModule, MAKEINTRESOURCE(iconId), IMAGE_ICON, 0, 0, LR_DEFAULTCOLOR));
					}

					delete pHelper;

					bResult = true;
				}
			}
		}

		delete [] pVerData;
	}

	return bResult;
}

#endif

#endif // DOPUS_PLUGIN_LEO_NO_COPYVERSION
